// -*- coding: utf-8 -*-
//
//  interpolation.h
//
//  Created by Deepansh J. Srivastava, Apr 11, 2019
//  Contact email = deepansh2012@gmail.com
//

#include "config.h"

extern int triangle_interpolation(double *freq1, double *freq2, double *freq3,
                                  double *amp, double *spec, int *points);
