// -*- coding: utf-8 -*-
//
//  powder_setup.h
//
//  Created by Deepansh J. Srivastava, Apr 11, 2019
//  Contact email = deepansh2012@gmail.com
//

#include "config.h"

extern void octahedron_averaging_setup(int nt, void *exp_I_alpha,
                                       void *exp_I_beta, double *amp);
